module casper.compiler {
   compile_file(Filename, Target) :-
      new():compile_file(Filename, Target).

   new(M) :-
      FreshVariableCounter = data.ref.backtrackable:new(0),
      module M [FreshVariableCounter] {
         fresh_var(X) :-
            X = data.atom:format('x~d', [data.number.(FreshVariableCounter++)]).

         sequence_to_list('position'(A, _)) =
            ! >>
            sequence_to_list(A).

         sequence_to_list((A, B)) =
            ! >>
            [A | sequence_to_list(B)].

         sequence_to_list(A) = [A].

         remove_position('position'(A, _)) =
            ! >>
            remove_position(A).

         remove_position(A) = A.

         binder(A, Pos) =
            remove_position(A) match (
               Atom = data.atom:check
            ->
               Atom
            ;
               casper.error:print(Pos, 'Binder expected') >>
               'error'
            ).

         decorate_type(Input, Output) :-
            module Decorate_type [Decorate_type] {
               do('position'(A, Pos), _Pos) =
                  ! >>
                  do(A, Pos).

               do('string', _Pos) =
                  ! >>
                  'string'.

               do('goal', _Pos) =
                  ! >>
                  'goal'.

               do('int', _Pos) =
                  ! >>
                  'int'.

               do('bool', _Pos) =
                  ! >>
                  'bool'.

               do('var'(X), Pos) =
                  ! >>
                  'var'(do(X, Pos)).

               do('[]'(X), Pos) =
                  ! >>
                  '[]'(do(X, Pos)).

               do('{}'(Record), Pos) =
                  ! >>
                  '{}'(data.list:map(sequence_to_list(Record)) {
                     get('position'(A, _)) =
                        get(A).

                     get(':'('position'(Field, _), Value)) =
                        (Field, Decorate_type:do(Value, Pos))
                  }).
         
               do(E, _Pos) =
                  io.err:print_endline('unimplemented decorate_type'(E)) >>
                  'none'.
            }:do(Input, 'unknown', Output).

         decorate(Input, Output) :-
            module Decorate [Decorate] {
               do('position'(A, Pos), _Pos, Env) =
                  ! >>
                  do(A, Pos, Env).

               extract_function(Decl, Name, Pos, Env, NewEnv, Body, NewBody) :-
                  (
                     Decl = 'position'(A, Pos')
                  ->
                     extract_function(A, Name, Pos', Env, NewEnv, Body, NewBody)
                  ;
                     data.atom:check(Decl)
                  ->
                     Name = Decl,
                     NewEnv = Env,
                     NewBody = Body
                  ;
                     data.term:check(Decl)
                  ->
                     [Name | Args] = data.term:decomposition(Decl),
                     decorate_args(Args, Env, Args', NewEnv),
                     Body = '->'(Args', NewBody, Pos)
                  ;
                     casper.error:print(Pos, 'Name or function expected'),
                     Name = 'error',
                     NewEnv = Env,
                     NewBody = Body
                  ).

               do('let'('position'('in'('position'(':='(Binder, Expr), _), Body), _)), Pos, Env) =
                  ! >>
                  (
                     extract_function(Binder, Binder', Pos, Env, NewEnv, Body', NewBody),
                     NewBody = do(Body, Pos, [Binder' | NewEnv])
                  ) >>
                  'let'(Binder', do(Expr, Pos, Env), Body', Pos).

               do('parameter'('position'(Name, _)), Pos, _Env) =
                  ! >>
                  'parameter'(Name, Pos).

               do('import'('position'(Name, _)), Pos, _Env) =
                  ! >>
                  'import'(Name, Pos).

               do('solve'(Expr), Pos, Env) =
                  ! >>
                  'builtin'('solve', [do(Expr, Pos, Env)], Pos).
         
               do('forall'('position'(','('position'('in'(Binder, Enum), _), Expr), _)), Pos, Env) =
                  ! >>
                  (Binder' = binder(Binder, Pos)) >>
                  'forall'(Binder', do(Enum, Pos, Env), do(Expr, Pos, [Binder' | Env]), Pos).

               do(':'(A, B), Pos, Env) =
                  ! >>
                  ':'(do(A, Pos, Env), decorate_type(B), Pos).

               decorate_arg('position'(A, _)) =
                  decorate_arg(A).

               decorate_arg(Atom) =
                  data.atom:check(Atom) >>
                  ! >>
                  Atom.

               decorate_args([], Env, [], Env).

               decorate_args([Arg | Args], Env, [Arg' | Args'], Env') :-
                  decorate_arg(Arg, Arg'),
                  decorate_args(Args, [Arg' | Env], Args', Env').

               do('->'(A, B), Pos, Env) =
                  ! >>
                  (Args = sequence_to_list(A)) >>
                  decorate_args(Args, Env, Args', Env') >>
                  '->'(Args', do(B, Pos, Env'), Pos).
         
               do('/\\'(A, B), Pos, Env) =
                  ! >>
                  'builtin'('/\\', [do(A, Pos, Env), do(B, Pos, Env)], Pos).
         
               do('\\/'(A, B), Pos, Env) =
                  ! >>
                  'builtin'('\\/', [do(A, Pos, Env), do(B, Pos, Env)], Pos).
         
               do('+'(A, B), Pos, Env) =
                  ! >>
                  'builtin'('+', [do(A, Pos, Env), do(B, Pos, Env)], Pos).
         
               do('-'(A, B), Pos, Env) =
                  ! >>
                  'builtin'('-', [do(A, Pos, Env), do(B, Pos, Env)], Pos).
         
               do('..'(A, B), Pos, Env) =
                  ! >>
                  'builtin'('..', [do(A, Pos, Env), do(B, Pos, Env)], Pos).
         
               do(''(A, 'position'('[]'(B), _)), Pos, Env) =
                  ! >>
                  'builtin'('nth', [do(A, Pos, Env), do(B, Pos, Env)], Pos).
         
               do('!='(A, B), Pos, Env) =
                  ! >>
                  'builtin'('!=', [do(A, Pos, Env), do(B, Pos, Env)], Pos).
         
               do('<-'(A, B), Pos, Env) =
                  ! >>
                  'builtin'('<-', [do(A, Pos, Env), do(B, Pos, Env)], Pos).
         
               do('<='(A, B), Pos, Env) =
                  ! >>
                  'builtin'('<=', [do(A, Pos, Env), do(B, Pos, Env)], Pos).
         
               do('inf'(Expr), Pos, Env) =
                  ! >>
                  'builtin'('inf', [do(Expr, Pos, Env)], Pos).
         
               do('sup'(Expr), Pos, Env) =
                  ! >>
                  'builtin'('sup', [do(Expr, Pos, Env)], Pos).
         
               do('print'(Expr), Pos, Env) =
                  ! >>
                  'builtin'('print', [do(Expr, Pos, Env)], Pos).
         
               do('println', Pos, _Env) =
                  ! >>
                  'builtin'('println', [], Pos).
         
               do('labeling_up'(Expr), Pos, Env) =
                  ! >>
                  'builtin'('labeling_up', [do(Expr, Pos, Env)], Pos).
         
               do('ln', Pos, _Env) =
                  ! >>
                  'builtin'('ln', [], Pos).

               do('_', Pos, _Env) =
                  ! >>
                  '_'(Pos).

               do(X, Pos,  _Env) =
                  data.number:check(X) >>
                  ! >>
                  'int'(X, Pos).
         
               do(X, Pos, Env) =
                  data.atom:check(X) >>
                  ! >>
                  (
                     data.list:member(X, Env)
                  ->
                     'var'(X, Pos)
                  ;
                     io.err:print_endline(Env) >>
                     casper.error:print(Pos, data.atom:format('Unbound variable ~a', [X])) >>
                     'error'
                  ).
         
               do('string'(X), Pos, _Env) =
                  ! >>
                  'string'(X, Pos).

               do('[]'('position'('|'(Expr, 'position'('in'(Binder, Enum), _)), _)), Pos, Env) =
                  ! >>
                  (Binder' = binder(Binder, Pos)) >>
                  'map'(do(Expr, Pos, [Binder' | Env]), Binder', do(Enum, Pos, Env), Pos).

               do('{}'(Record), Pos, Env) =
                  ! >>
                  '{}'(data.list:map(sequence_to_list(Record)) [Env, Decorate, Pos] {
                     get(A) = get(A, Pos).

                     get('position'(A, Pos'), _Pos) =
                        (
                           X = get(A, Pos')
                        ->
                           X
                        ;
                           casper.error:print(Pos', 'Field definition excepted') >>
                           'error'
                        ).

                     get(':'(Field, Value), Pos') =
                        extract_function(Field, Field', Pos', Env, Env', Body', NewBody) >>
                        (NewBody = Decorate:do(Value, Pos', Env')) >>
                        (Field', Body').
                  }, Pos).
         
               do(E, _Pos, _Env) =
                  io.err:print_endline('unimplemented decorate'(E)) >>
                  'none'.
            }:do(Input, 'unknown', [], Output).

         eliminate_parameters(Input, Parameters, Output) :-
            ParametersRef = data.ref.backtrackable:new([]),
            module EliminateParameters [ParametersRef, EliminateParameters] {
               get_parameter(Name, Param) :-
                  Parameters = ParametersRef:get,
                  (
                     data.list:member((Param, Name), Parameters)
                  ->
                     true
                  ;
                     Param = fresh_var,
                     ParametersRef:set([(Param, Name) | Parameters])
                  ).

               do('let'(Binder, Expr, Body, Pos)) =
                  ! >>
                  'let'(Binder, do(Expr), do(Body), Pos).
         
               do('parameter'(Name, Pos)) =
                  ! >>
                  (
                     Name = 'string'(VarName)
                  ->
                     'var'(get_parameter(VarName), Pos)
                  ).
         
               do('forall'(Binder, Range, Expr, Pos)) =
                  ! >>
                  'forall'(Binder, do(Range), do(Expr), Pos).

               do('builtin'(Builtin, Args, Pos)) =
                  ! >>
                  'builtin'(Builtin, data.list:iter(Args, EliminateParameters), Pos).
         
               do(':'(A, B, Pos)) =
                  ! >>
                  ':'(do(A), B, Pos).
         
               do('->'(A, B, Pos)) =
                  ! >>
                  '->'(A, do(B), Pos).

               do(A = ('_'(_Pos) ; 'int'(_I, _Pos) ; 'var'(_X, _Pos) ; 'string'(_S, _Pos))) =
                  ! >>
                  A.

               do('map'(Expr, Binder, Enum, Pos)) =
                  ! >>
                  'map'(do(Expr), Binder, do(Enum), Pos).

               do('{}'(Record, Pos)) =
                  ! >>
                  '{}'(
                     data.list:map(Record) [EliminateParameters] {
                        get((Label, Value)) =
                           (Label, EliminateParameters:do(Value)).
                     },
                     Pos
                  ).

               do(E, 'none') :-
                  data.term:functor(E, F, N),
                  io.err:print_endline('unimplemented eliminate_parameters'('/'(F, N))).
         
            }:do(Input, Output),
            Parameters = ParametersRef:get().

         eliminate_imports(Input, Output) :-
            ImportsRef = data.ref.backtrackable:new([]),
            module EliminateImports [ImportsRef, EliminateImports] {
               get_import(Name, Import) :-
                  Imports = ImportsRef:get,
                  (
                     data.list:member((Import, Name), Imports)
                  ->
                     true
                  ;
                     Import = fresh_var,
                     ImportsRef:set([(Import, Name) | Imports])
                  ).

               do('let'(Binder, Expr, Body, Pos)) =
                  ! >>
                  'let'(Binder, do(Expr), do(Body), Pos).
         
               do('import'(Name, Pos)) =
                  ! >>
                  (
                     Name = 'string'(VarName)
                  ->
                     'var'(get_import(VarName), Pos)
                  ).
         
               do('forall'(Binder, Range, Expr, Pos)) =
                  ! >>
                  'forall'(Binder, do(Range), do(Expr), Pos).

               do('builtin'(Builtin, Args, Pos)) =
                  ! >>
                  'builtin'(Builtin, data.list:iter(Args, EliminateImports), Pos).
         
               do(':'(A, B, Pos)) =
                  ! >>
                  ':'(do(A), B, Pos).
         
               do('->'(A, B, Pos)) =
                  ! >>
                  '->'(A, do(B), Pos).

               do(A = ('_'(_Pos) ; 'int'(_I, _Pos) ; 'var'(_X, _Pos) ; 'string'(_S, _Pos) ; 'parameter'(_S, _Pos))) =
                  ! >>
                  A.

               do('map'(Expr, Binder, Enum, Pos)) =
                  ! >>
                  'map'(do(Expr), Binder, do(Enum), Pos).

               do('{}'(Record, Pos)) =
                  ! >>
                  '{}'(
                     data.list:map(Record) [EliminateImports] {
                        get((Label, Value)) =
                           (Label, EliminateImports:do(Value)).
                     },
                     Pos
                  ).

               do(E, 'none') :-
                  data.term:functor(E, F, N),
                  io.err:print_endline('unimplemented eliminate_imports'('/'(F, N))).
         
            }:do(Input, Output0),
            Output = data.list:fold_left(ImportsRef:get(), Output0) [] {
               do((Import, Name), Output) =
                  casper.parser:parse_file(Name, Ast) >>
                  'let'(Import, eliminate_imports(decorate(Ast)), Output, 'dummy').                  
            }.

         remove_pos(Input, Output) :-
            module RemovePos [RemovePos] {
               do('let'(Binder, Expr, Body, _Pos)) =
                  ! >>
                  'let'(Binder, do(Expr), do(Body)).
         
               do('parameter'(Name, _Pos)) =
                  ! >>
                  'parameter'(Name).
         
               do('forall'(Binder, Range, Expr, _Pos)) =
                  ! >>
                  'forall'(Binder, do(Range), do(Expr)).

               do('builtin'(Builtin, Args, _Pos)) =
                  ! >>
                  'builtin'(Builtin, data.list:iter(Args, RemovePos)).
         
               do(':'(A, B, _Pos)) =
                  ! >>
                  ':'(do(A), B).

               do('_'(_Pos)) =
                  ! >>
                  '_'.

               do('int'(I, _Pos)) =
                  ! >>
                  'int'(I).

               do('string'(S, _Pos)) =
                  ! >>
                  'string'(S).

               do('var'(X, _Pos)) =
                  ! >>
                  'var'(X).

               do('map'(Expr, Binder, Enum, _Pos)) =
                  ! >>
                  'map'(do(Expr), Binder, do(Enum)).

               do('{}'(Record, _Pos)) =
                  ! >>
                  '{}'(
                     data.list:map(Record) [EliminateParameters] {
                        get((Label, Value)) =
                           (Label, EliminateParameters:do(Value)).
                     }
                  ).

               do(E, 'none') :-
                  data.term:functor(E, F, N),
                  io.err:print_endline('unimplemented remove_pos'('/'(F, N))).
         
            }:do(Input, Output).

         desugar(Input, Output) :-
            module Desugar [Desugar] {
               do('let'(Binder, Expr, Body, Pos)) =
                  ! >>
                  'let'(Binder, do(Expr), do(Body), Pos).
         
               do('parameter'(Name, Pos)) =
                  ! >>
                  'parameter'(Name, Pos).

               do('forall'(Binder, Range, Expr, Pos)) =
                  ! >>
                  'forall'(Binder, do(Range), do(Expr), Pos).

               do('builtin'('<=', [A, ':'('builtin'('<=', [B, C], _), _, _)], Pos)) =
                  ! >>
                  'builtin'('/\\', ['builtin'('<=', [do(A), do(B)], _), 'builtin'('<=', [do(B), do(C)], _)], Pos).

               do('builtin'('<=', [A, 'builtin'('<=', [B, C], _)], Pos)) =
                  ! >>
                  'builtin'('/\\', ['builtin'('<=', [A, B], _), 'builtin'('<=', [B, C], _)], Pos).

               do('builtin'(Builtin, Args, Pos)) =
                  ! >>
                  'builtin'(Builtin, data.list:iter(Args, Desugar), Pos).
         
               do(':'(A, B, Pos)) =
                  ! >>
                  ':'(do(A), B, Pos).
         
               do('var'(X, Pos)) =
                  ! >>
                  'var'(X, Pos).
         
               do('int'(X, Pos)) =
                  ! >>
                  'int'(X, Pos).
         
               do('string'(X, Pos)) =
                  ! >>
                  'string'(X, Pos).

               do('_'(Pos)) =
                  ! >>
                  '_'(Pos).

               do('map'(Expr, Binder, Enum, Pos)) =
                  ! >>
                  'map'(do(Expr), Binder, do(Enum), Pos).

               do('{}'(Record, Pos)) =
                  ! >>
                  '{}'(
                     data.list:map(Record) [Desugar] {
                        get((Label, Value)) =
                           (Label, Desugar:do(Value)).
                     },
                     Pos
                  ).

               do(E, 'none') :-
                  io.err:print_endline('unimplemented desugar'(E)).
         
            }:do(Input, Output).

         infer_types(Input, Parameters, Output) :-
            module [] {
               get_type(':'(_, Type, _)) = Type.

               expect_type(':'(_, Type, Pos), Expected, Error) :-
                  (
                     Type = Expected
                  ->
                     true
                  ;
                     casper.error:print(Pos, Error)
                  ).

               get_item_type(Expr, ItemType) :-
                  expect_type(Expr, '[]'(ItemType), 'List expected').

               do('let'(Binder, Expr, Body, Pos), Env) =
                  ! >>
                  (NewExpr = do(Expr, Env)) >>
                  (NewBody = do(Body, [':'(Binder, get_type(NewExpr)) | Env])) >>
                  ':'('let'(Binder, NewExpr, NewBody, Pos), get_type(NewBody), Pos).
         
               do('parameter'(Name, Pos), _) =
                  ! >>
                  ':'('parameter'(Name, Pos), _, Pos).

               do('forall'(Binder, Range, Expr, Pos), Env) =
                  ! >>
                  (NewRange = do(Range, Env)) >>
                  (ItemType = get_item_type(NewRange)) >>
                  (NewExpr = do(Expr, [':'(Binder, ItemType) | Env])) >>
                  ':'(
                     'forall'(Binder, NewRange, NewExpr, Pos),
                     get_type(NewExpr),
                     Pos
                  ).

               builtin_type('solve', ['{}'([('constraint', 'var'('bool')), ('search', 'goal'), ('output', 'goal')])], 'goal').
               builtin_type('..', ['int', 'int'], '[]'('int')).
               builtin_type('+', ['int', 'int'], 'int').
               builtin_type('+', ['var'('int'), 'var'('int')], 'var'('int')).
               builtin_type('+', ['string', 'string'], 'string').
               builtin_type('-', ['int', 'int'], 'int').
               builtin_type('-', ['var'('int'), 'var'('int')], 'var'('int')).
               builtin_type('*', ['int', 'int'], 'int').
               builtin_type('*', ['var'('int'), 'var'('int')], 'var'('int')).
               builtin_type('/', ['int', 'int'], 'int').
               builtin_type('/', ['var'('int'), 'var'('int')], 'var'('int')).
               builtin_type('/\\', ['bool', 'bool'], 'bool').
               builtin_type('/\\', ['var'('bool'), 'var'('bool')], 'var'('bool')).
               builtin_type('/\\', ['goal', 'goal'], 'goal').
               builtin_type('\\/', ['bool', 'bool'], 'bool').
               builtin_type('\\/', ['var'('bool'), 'var'('bool')], 'var'('bool')).
               builtin_type('!=', ['int', 'int'], 'bool').
               builtin_type('!=', ['var'('int'), 'var'('int')], 'var'('bool')).
               builtin_type('<=', ['int', 'int'], 'bool').
               builtin_type('<=', ['var'('int'), 'var'('int')], 'var'('bool')).
               builtin_type('<-', ['var'('int'), '[]'('int')], 'goal').
               builtin_type('inf', ['var'('int')], 'int').
               builtin_type('sup', ['var'('int')], 'int').
               builtin_type('nth', ['[]'(A), 'int'], A).
               builtin_type('print', ['int'], 'goal').
               builtin_type('print', ['string'], 'goal').
               builtin_type('println', [], 'goal').
               builtin_type('ln', [], 'string').

               subtype(A, A).

               subtype('bool', 'int').

               subtype('bool', 'var'('bool')).

               subtype('bool', 'var'('int')).

               subtype('bool', 'string').

               subtype('int', 'var'('int')).

               subtype('int', 'string').

               subtype('var'('bool'), 'var'('int')).

               subtype('[]'(A), '[]'(B)) :-
                  subtype(A, B).

               subtype('{}'(A), '{}'(B)) :-
                  \+ (
                     data.list:member((F, TB), B),
                     \+ (
                        data.list:member((F, TA), A),
                        subtype(TA, TB)
                     )
                  ).

               get_typed_args(Args, Env) = data.list:map(Args) {
                  get(Arg) = do(Arg, Env)
               }.

               get_arg_types(TypedArgs) = data.list:map(TypedArgs) {
                  get(Arg) = get_type(Arg)
               }.

               get_builtin_result_type(Builtin, ArgTypes, Pos, ResultType) :-
                  (
                     builtin_type(Builtin, ArgTypes', ResultType),
                     data.list:iter(data.list:zip(ArgTypes, ArgTypes')) {
                        do((A, B)) :- subtype(A, B)
                     }
                  ->
                     true
                  ;
                     casper.error:print(Pos, data.atom:format('Invalid argument types ~q for ~q', [ArgTypes, Builtin]))
                  ).

               do('builtin'(Builtin, Args, Pos), Env) =
                  ! >> 
                  (TypedArgs = get_typed_args(Args, Env)) >>
                  (ArgTypes = get_arg_types(TypedArgs)) >>
                  (ResultType = get_builtin_result_type(Builtin, ArgTypes, Pos)) >>
                  ':'('builtin'(Builtin, TypedArgs, Pos), ResultType, Pos).

               do(':'(A, B, Pos), Env) =
                  ! >>
                  do(A, Env) match (
                     ':'(A, AType, _)
                  ->
                     (
                        subtype(AType, B)
                     ->
                        true
                     ;
                        casper.warning:print(Pos, data.atom:format('This expression has type ~q which is not convertible into ~q', [AType, B]))
                     ) >>
                     ':'(A, B, Pos)
                  ).

               do('var'(X, Pos), Env) =
                  ! >>
                  (
                     Type = (
                        data.list:member(':'(X, Type), Env)
                     ->
                        Type
                     ;
                        io.err:print_endline('unbound variable'(X)) >>
                        'error'
                     )
                  ) >>
                  ':'('var'(X, Pos), Type, Env).

               do('int'(X, Pos), _Env) =
                  ! >>
                  ':'('int'(X, Pos), 'int', Pos).

               do('string'(X, Pos), _Env) =
                  ! >>
                  ':'('string'(X, Pos), 'string', Pos).

               do('_'(Pos), _Env) =
                  ! >>
                  ':'('_'(Pos), 'var'('int'), Pos).

               do('map'(Expr, Binder, Enum, Pos), Env) =
                  ! >>
                  (NewEnum = do(Enum, Env)) >>
                  (NewExpr = do(Expr, [':'(Binder, get_item_type(NewEnum)) | Env])) >>
                  ':'(
                     'map'(NewExpr, Binder, NewEnum, Pos),
                     '[]'(get_type(NewExpr)),
                     Pos
                  ).

               do('{}'(Record, Pos), Env) =
                  ! >> (
                     TypedRecord = data.list:map(Record) [Env] {
                        get((Label, Value)) =
                           (Label, do(Value, Env)).
                     },
                     RecordType = data.list:map(TypedRecord) [] {
                        get((Label, Value)) =
                           (Label, get_type(Value)).
                     }
                  ) >>
                  ':'(
                     '{}'(TypedRecord, Pos),
                     '{}'(RecordType),
                     Pos
                  ).

               do(E, _Env) =
                  io.err:print_endline('unimplemented type_infer'(E)) >>
                  'none'.
            }:do(
               Input,
               data.list:map(Parameters) {
                  get((Param, _)) = ':'(Param, 'int').
               },
               Output
            ).

         standard_strategies(Input, Output) :-
            module StandardStrategies [StandardStrategies] {
               do(':'(Expr, Type, Pos)) =
                  ! >>
                  ':'(do(Expr), Type, Pos).

               do('let'(Binder, Expr, Body, Pos)) =
                  ! >>
                  'let'(Binder, do(Expr), do(Body), Pos).
         
               do('forall'(Binder, Range, Expr, Pos)) =
                  ! >>
                  'forall'(Binder, do(Range), do(Expr), Pos).
         
               do('parameter'(Name, Pos)) =
                  ! >>
                  'parameter'(Name, Pos).

               do('builtin'('<-', [A, ':'('builtin'('..', [':'('builtin'('inf', [B], _), 'int', _), ':'('builtin'('sup', [C], _), 'int', _)], _), '[]'('int'), _)], Pos)) =
                  (remove_pos(A) = remove_pos(B) = remove_pos(C)) >>
                  ! >>
                  'builtin'('labeling_up', [A], Pos).

               do('builtin'(Builtin, Args, Pos)) =
                  ! >>
                  'builtin'(Builtin, data.list:map(Args) { get(X) = do(X) }, Pos).
         
               do(A = ('var'(_X, _Pos) ; 'int'(_I, _Pos) ; 'string'(_S, _Pos) ; '_'(_Pos))) =
                  ! >>
                  A.
         
               do('map'(Expr, Binder, Enum, Pos)) =
                  ! >>
                  'map'(do(Expr), Binder, do(Enum), Pos).

               do('{}'(Record, Pos)) =
                  ! >>
                  '{}'(
                     data.list:map(Record) [StandardStrategies] {
                        get((Label, Value)) =
                           (Label, StandardStrategies:do(Value)).
                     },
                     Pos
                  ).

               do(E, 'none') :-
                  io.err:print_endline('unimplemented standard_strategies'(E)).
         
            }:do(Input, Output).

         compile_file(SourcePath, Target) :-
            casper.parser:parse_file(SourcePath, Ast),
            data.path.(
               split(SourcePath, BasePath, SourceFilename),
               chop_suffix(SourceFilename, SourceBase)
            ),
            (
               Target = 'minisat/java'
            ->
               ClassName = data.atom:capitalize(SourceBase),
               TargetFilename = data.atom:concat(ClassName, '.java'),
               TargetPath = data.path:concat(BasePath, TargetFilename),
               Expr = Ast
                 | decorate
                 | eliminate_imports
                 | eliminate_parameters(Parameters)
                 | desugar
                 | infer_types(Parameters)
                 | standard_strategies,
               casper.codegen.minisat:generate(ClassName, Parameters, Expr, TargetPath),
               sys.process:spawn('javac', [TargetFilename])
            ;
               Target = 'choco/java'
            ->
               ClassName = data.atom:capitalize(SourceBase),
               TargetFilename = data.atom:concat(ClassName, '.java'),
               TargetPath = data.path:concat(BasePath, TargetFilename),
               Expr = Ast
                 | decorate
                 | eliminate_imports
                 | eliminate_parameters(Parameters)
                 | desugar
                 | infer_types(Parameters)
                 | standard_strategies,
               casper.codegen.choco:generate(ClassName, Parameters, Expr, TargetPath),
               sys.process:spawn('javac', [TargetFilename])
            ;
               Target = 'typed/casper'
            ->
               TargetFilename = data.atom:concat(SourceBase, '_typed.csp'),
               TargetPath = data.path:concat(BasePath, TargetFilename),
               Expr = Ast
                 | decorate
                 | eliminate_imports
                 | desugar
                 | infer_types([]),
               casper.codegen.casper:generate(Expr, TargetPath)
            ;
               throw('error'(data.atom:format('Unknown target ~a', [Target])))
            ).
      }.

/*      casper.parser:parse_file(Filename, Ast),
      io.std:print_endline(Ast).*/
}